#Imports

import csv



#Load data
#	Use 1931 standard to ease any comparisons with previous work, which is more-likely to use the
#	older standard.

def load_csv(path,skiprows):
	with open(path,"r",encoding="utf-8") as csv_file:
		reader = csv.reader(csv_file, delimiter=",")
		rows = [ row for row in reader ]

	#Throw out the headers and everything below 380nm
	rows = rows[skiprows:]

	#Transpose into columns and parse floating-point
	cols = [ [] for i in range(len(rows[0])) ]
	for row in rows:
		i = 0
		for elem in row:
			if elem=="": elem=0.0
			else:        elem=float(elem)
			cols[i].append(elem)
			i += 1

	return cols

stdobs      = load_csv("../simple-spectral/data/cie1931-xyzbar-380+5+780.csv",0)
illuminants = load_csv("illuminants.csv",2+16)
basis       = load_csv("../simple-spectral/data/cie1931-basis-bt709-380+5+780.csv",0)

stdobs = { "x̄":stdobs[0], "ȳ":stdobs[1], "z̄":stdobs[2] }
cie_x̄ = stdobs["x̄"]
cie_ȳ = stdobs["ȳ"]
cie_z̄ = stdobs["z̄"]

N = len(cie_x̄)

illuminants = {
	"A"      : illuminants[ 1],
	"D65"    : illuminants[ 2],
	"C"      : illuminants[ 3],
	"D50"    : illuminants[ 4],
	"D55"    : illuminants[ 5],
	"D75"    : illuminants[ 6],

	"FL1"    : illuminants[ 7],
	"FL2"    : illuminants[ 8],
	"FL3"    : illuminants[ 9],
	"FL4"    : illuminants[10],
	"FL5"    : illuminants[11],
	"FL6"    : illuminants[12],
	"FL7"    : illuminants[13],
	"FL8"    : illuminants[14],
	"FL9"    : illuminants[15],
	"FL10"   : illuminants[16],
	"FL11"   : illuminants[17],
	"FL12"   : illuminants[18],

	"FL3.1"  : illuminants[19],
	"FL3.2"  : illuminants[20],
	"FL3.3"  : illuminants[21],
	"FL3.4"  : illuminants[22],
	"FL3.5"  : illuminants[23],
	"FL3.6"  : illuminants[24],
	"FL3.7"  : illuminants[25],
	"FL3.8"  : illuminants[26],
	"FL3.9"  : illuminants[27],
	"FL3.10" : illuminants[28],
	"FL3.11" : illuminants[29],
	"FL3.12" : illuminants[30],
	"FL3.13" : illuminants[31],
	"FL3.14" : illuminants[32],
	"FL3.15" : illuminants[33],

	"HP1"    : illuminants[34],
	"HP2"    : illuminants[35],
	"HP3"    : illuminants[36],
	"HP4"    : illuminants[37],
	"HP5"    : illuminants[38],

	"E"      : [ 1 for i in range(N) ]
}

basis = { "Sr":basis[0], "Sg":basis[1], "Sb":basis[2] }
S_r = basis["Sr"]
S_g = basis["Sg"]
S_b = basis["Sb"]



#Useful functions

def lrgb_to_srgb(lrgb):
	srgb_int = [None,None,None]
	for i in range(3):
		if lrgb[i] < 0.0031308:
			srgb_int[i] = 12.92 * lrgb[i]
		else:
			srgb_int[i] = 1.055*( lrgb[i] ** (1.0/2.4) ) - 0.055
		srgb_int[i] = int(round( srgb_int[i] * 255.0 ))
		if   srgb_int[i]<  0: srgb_int[i]=  0
		elif srgb_int[i]>255: srgb_int[i]=255
	return srgb_int
def srgb_to_lrgb(srgb_int):
	lrgb = [None,None,None]
	for i in range(3):
		val = srgb_int[i] / 255.0
		if val < 0.04045:
			lrgb[i] = val / 12.92
		else:
			lrgb[i] = ( (val+0.055) / 1.055 )**2.4
	return lrgb

def lerp( val0,val1, t ):
	return val0*(1.0-t) + val1*t
def lerp_srgb( srgb0_int,srgb1_int, t ):
	#This version is probably best (interpolate in linear RGB)
	lrgb0 = srgb_to_lrgb(srgb0_int)
	lrgb1 = srgb_to_lrgb(srgb1_int)
	lrgb = [ lerp( lrgb0[i],lrgb1[i], t ) for i in range(3) ]
	srgb_int = lrgb_to_srgb(lrgb)
	#This version is what [Jakob and Hanika 2019] appear to do in Figure 10 (interpolate in sRGB).
	##srgb_int = [
	##	int(round(lerp( srgb0_int[i],srgb1_int[i], t )))
	##	for i in range(3)
	##]
	return srgb_int

def flux_to_XYZ(spectral_flux):
	return (
		sum([ cie_x̄[i] * spectral_flux[i] for i in range(N) ]),
		sum([ cie_ȳ[i] * spectral_flux[i] for i in range(N) ]),
		sum([ cie_z̄[i] * spectral_flux[i] for i in range(N) ])
	)

#	sRGB triple illumination
def illuminate_color( srgb_int, name_illuminant ):
	r,g,b = srgb_to_lrgb(srgb_int)

	reflectance = [ S_r[i]*r + S_g[i]*g + S_b[i]*b for i in range(N) ]

	illuminant = illuminants[name_illuminant]
	scaled_radiance = [ illuminant[i] * reflectance[i] for i in range(N) ]

	scaled_flux = scaled_radiance #Assume flat-field correction

	X, Y, Z = flux_to_XYZ(scaled_flux)

	Y_illum = flux_to_XYZ(illuminant)[1]

	lrgb = [
		(  3.24083022913212560*X - 1.53731690356267480*Y - 0.49858926602032710*Z ) / Y_illum,
		( -0.96922932087485440*X + 1.87593979409188670*Y + 0.04155444365280374*Z ) / Y_illum,
		(  0.05564528732689767*X - 0.20403272019862467*Y + 1.05726045921105550*Z ) / Y_illum
	]

	srgb_int = lrgb_to_srgb(lrgb)

	return srgb_int

#	Gradient illumination
def illuminate_gradient( gradient, name_illuminant, path_output ):
	srgb0_int, srgb1_int = gradient
	import os, sys
	with open(os.devnull, "w") as f:
		oldstdout=sys.stdout; sys.stdout=f; import pygame; sys.stdout=oldstdout

	W = 256
	surf = pygame.Surface((W,2))
	for i in range(W):
		srgb_int_in  = lerp_srgb( srgb0_int,srgb1_int, float(i)/float(W-1) )
		srgb_int_out = illuminate_color( srgb_int_in, name_illuminant )
		surf.set_at((i,0),srgb_int_in )
		surf.set_at((i,1),srgb_int_out)

	##surf = pygame.transform.scale(surf,(W,100))

	pygame.image.save(surf,path_output)



#Illuminate some gradients

gradient_bluepink  = ( (0,0,255), (255,0,255) )
gradient_redwhite  = ( (255,0,0), (255,255,255) )
gradient_cyangreen = ( (0,127,255), (0,255,0) )

for illumname in ["A","D65","D50","E","FL1","FL2","FL3","FL3.1","FL3.2","FL3.3","HP1","HP2","HP3"]:
	for gradient,gradname in [(gradient_bluepink,"bluepink"),(gradient_redwhite,"redwhite"),(gradient_cyangreen,"cyangreen")]:
		illuminate_gradient( gradient, illumname, illumname+"_"+gradname+".png" )

##illuminate_gradient( ((0,0,255),(255,0,255)), "E",   "E_bluepink.png"   )
##illuminate_gradient( ((0,0,255),(255,0,255)), "FL1", "FL1_bluepink.png" )
##illuminate_gradient( ((0,0,255),(255,0,255)), "FL2", "FL2_bluepink.png" )
##illuminate_gradient( ((0,0,255),(255,0,255)), "FL3", "FL3_bluepink.png" )
